/*
 * Copyright (C) 2001-2003  Simon Baldwin (simon_baldwin@yahoo.com)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#ifndef IFP_INTERNAL_H
#define IFP_INTERNAL_H


/* Convenience macro for packing version numbers. */
#define	IFP_VERSION(MAJOR,MIDDLE,MINOR)					\
				(((MAJOR) << 16) | ((MIDDLE) << 8) | (MINOR))


/* Engine header version definition. */
#define	IFP_HEADER_VERSION_0_2_0	IFP_VERSION(0,2,0)


/*
 * IF engine header structure.  Each engine exports one of these, and it
 * defines a few things about the engine contained in the plugin.
 */
struct ifp_header {
	int	version;		/* Header version used by the engine. */
	char	*build_timestamp;	/* Date and time of plugin build */

	char	*engine_type;		/* Data type the engine runs. */
	char	*engine_name;		/* Unique engine name. */
	char	*engine_version;	/* Engine's version number. */

	char	*blorb_pattern;		/* If Blorb-capable, the exec type. */

	int	acceptor_offset;	/* Offset in file to "magic". */
	int	acceptor_length;	/* Length of "magic". */
	char	*acceptor_pattern;	/* Regular expression for "magic". */

	char	*author_name;		/* Interpreter author's name. */
	char	*author_email;		/* Interpreter author's email. */
	char	*engine_home_url;	/* Any URL regarding the interpreter. */

	char	*builder_name;		/* Engine porter/builder's name. */
	char	*builder_email;		/* Engine porter/builder's email. */

	char	*engine_description;	/* Miscellaneous engine information. */
	char	*engine_copyright;	/* Engine's copyright information. */
};

/* Skip all except the header definition for IFP plugin headers. */
#ifndef IFP_HEADER_ONLY

/* Inclusions required for type and other definitions. */
#include <sys/types.h>
#include <stdio.h>

/* Glk dispatch and Blorb definitions. */
#include <gi_dispa.h>
#include <gi_blorb.h>

/* Required IFP main definitions. */
#include "ifp.h"


/*
 * The environment variable used as the path for loading plugins.  This is
 * an approximate analog to LD_LIBRARY_PATH.
 */
#define	IFP_PLUGIN_PATH		"IF_PLUGIN_PATH"


/* Libc interface version definition. */
#define	IFP_LIBC_VERSION_0_3_0	IFP_VERSION(0,3,0)

/*
 * Library call proxy interface.  This is a vector of functions which would
 * normally be resolved by libc, but which for one reason or another we want
 * to trap in a plugin and bring back to the main program for some additional
 * processing.  Typically, this would be something like malloc(), so that
 * the program that loads the plugin can track what it uses, and garbage-
 * collect when the plugin completes.
 */
struct ifp_libc_interface {
	int	version;

	void	*(*malloc) (size_t);
	void	*(*calloc) (size_t, size_t);
	void	*(*realloc) (void*,size_t);
	char	*(*strdup) (const char*);
	char	*(*getcwd) (char *, size_t);
	void	(*free) (void*);

	int	(*open_2) (const char*, int);
	int	(*open_3) (const char*, int, mode_t);
	int	(*close) (int);
	int	(*creat) (const char*, mode_t);
	int	(*dup) (int);
	int	(*dup2) (int, int);

	FILE	*(*fopen) (const char*, const char*);
	FILE	*(*fdopen) (int, const char*);
	FILE	*(*freopen) (const char*, const char*, FILE*);
	int	(*fclose) (FILE*);
};


/*
 * Prototypes for libc interception functions.  These are in place only
 * for the plugin handler library code - a caller of plugin functions
 * should not need to worry about them.
 */
extern void	ifp_libc_intercept_exit (int status);
extern void	*ifp_libc_intercept_malloc (size_t size);
extern void	*ifp_libc_intercept_calloc (size_t nmemb, size_t size);
extern void	*ifp_libc_intercept_realloc (void *ptr, size_t size);
extern char	*ifp_libc_intercept_strdup (const char *s);
extern char	*ifp_libc_intercept_getcwd (char *buf, size_t size);
extern void	ifp_libc_intercept_free (void *ptr);

extern int	ifp_libc_intercept_open_2 (const char *pathname, int flags);
extern int	ifp_libc_intercept_open_3 (const char *pathname,
						int flags, mode_t mode);
extern int	ifp_libc_intercept_close (int fd);
extern int	ifp_libc_intercept_creat (const char *pathname, mode_t mode);
extern int	ifp_libc_intercept_dup (int oldfd);
extern int	ifp_libc_intercept_dup2 (int oldfd, int newfd);
extern FILE	*ifp_libc_intercept_fopen (const char *path, const char *mode);
extern FILE	*ifp_libc_intercept_fdopen (int filedes, const char *mode);
extern FILE	*ifp_libc_intercept_freopen (const char *path, const char *mode,
						FILE *stream);
extern int	ifp_libc_intercept_fclose (FILE *stream);


/*
 * Glk interface version definition.  Note that this is the interface
 * structure version number, not the version number of any used Glk library.
 */
#define	IFP_GLK_VERSION_0_3_1	IFP_VERSION(0,3,1)

/*
 * Glk proxy interface.  This is a vector of glk functions.  A plugin is
 * supplied with a copy of this table that the main program has filled out.
 * The proxy interface will then take each Glk call from the plugin, and
 * handle it by passing to the relevant function from this vector.
 *
 * There is, simply, one entry in the vector for each Glk function that
 * exists in the supported version of Glk.
 */
struct ifp_glk_interface {
	int	version;

	/* Base Glk functions. */
	void	(*glk_exit) (void);
	void	(*glk_set_interrupt_handler) (void (*func)(void));
	void	(*glk_tick) (void);

	glui32	(*glk_gestalt) (glui32 sel, glui32 val);
	glui32	(*glk_gestalt_ext) (glui32 sel, glui32 val, glui32 *arr,
				glui32 arrlen);

	unsigned char	(*glk_char_to_lower) (unsigned char ch);
	unsigned char	(*glk_char_to_upper) (unsigned char ch);

	winid_t	(*glk_window_get_root) (void);
	winid_t (*glk_window_open) (winid_t split, glui32 method, glui32 size,
				glui32 wintype, glui32 rock);
	void	(*glk_window_close) (winid_t win, stream_result_t *result);
	void	(*glk_window_get_size) (winid_t win, glui32 *widthptr,
				glui32 *heightptr);
	void	(*glk_window_set_arrangement) (winid_t win, glui32 method,
				glui32 size, winid_t keywin);
	void	(*glk_window_get_arrangement) (winid_t win, glui32 *methodptr,
				glui32 *sizeptr, winid_t *keywinptr);
	winid_t	(*glk_window_iterate) (winid_t win, glui32 *rockptr);
	glui32	(*glk_window_get_rock) (winid_t win);
	glui32	(*glk_window_get_type) (winid_t win);
	winid_t	(*glk_window_get_parent) (winid_t win);
	winid_t	(*glk_window_get_sibling) (winid_t win);
	void	(*glk_window_clear) (winid_t win);
	void	(*glk_window_move_cursor) (winid_t win, glui32 xpos,
				glui32 ypos);

	strid_t	(*glk_window_get_stream) (winid_t win);
	void	(*glk_window_set_echo_stream) (winid_t win, strid_t str);
	strid_t	(*glk_window_get_echo_stream) (winid_t win);
	void	(*glk_set_window) (winid_t win);

	strid_t	(*glk_stream_open_file) (frefid_t fileref, glui32 fmode,
				glui32 rock);
	strid_t	(*glk_stream_open_memory) (char *buf, glui32 buflen,
				glui32 fmode, glui32 rock);
	void	(*glk_stream_close) (strid_t str, stream_result_t *result);
	strid_t	(*glk_stream_iterate) (strid_t str, glui32 *rockptr);
	glui32	(*glk_stream_get_rock) (strid_t str);
	void	(*glk_stream_set_position) (strid_t str, glsi32 pos,
				glui32 seekmode);
	glui32	(*glk_stream_get_position) (strid_t str);
	void	(*glk_stream_set_current) (strid_t str);
	strid_t	(*glk_stream_get_current) (void);

	void	(*glk_put_char) (unsigned char ch);
	void	(*glk_put_char_stream) (strid_t str, unsigned char ch);
	void	(*glk_put_string) (char *s);
	void	(*glk_put_string_stream) (strid_t str, char *s);
	void	(*glk_put_buffer) (char *buf, glui32 len);
	void	(*glk_put_buffer_stream) (strid_t str, char *buf, glui32 len);
	void	(*glk_set_style) (glui32 styl);
	void	(*glk_set_style_stream) (strid_t str, glui32 styl);

	glsi32	(*glk_get_char_stream) (strid_t str);
	glui32	(*glk_get_line_stream) (strid_t str, char *buf, glui32 len);
	glui32	(*glk_get_buffer_stream) (strid_t str, char *buf, glui32 len);

	void	(*glk_stylehint_set) (glui32 wintype, glui32 styl, glui32 hint,
				glsi32 val);
	void	(*glk_stylehint_clear) (glui32 wintype, glui32 styl,
				glui32 hint);
	glui32	(*glk_style_distinguish) (winid_t win, glui32 styl1,
				glui32 styl2);
	glui32	(*glk_style_measure) (winid_t win, glui32 styl, glui32 hint,
				glui32 *result);

	frefid_t	(*glk_fileref_create_temp) (glui32 usage, glui32 rock);
	frefid_t	(*glk_fileref_create_by_name) (glui32 usage, char *name,
				glui32 rock);
	frefid_t	(*glk_fileref_create_by_prompt) (glui32 usage,
				glui32 fmode, glui32 rock);
	frefid_t	(*glk_fileref_create_from_fileref) (glui32 usage,
				frefid_t fref, glui32 rock);
	void		(*glk_fileref_destroy) (frefid_t fref);
	frefid_t	(*glk_fileref_iterate) (frefid_t fref, glui32 *rockptr);
	glui32	(*glk_fileref_get_rock) (frefid_t fref);
	void	(*glk_fileref_delete_file) (frefid_t fref);
	glui32	(*glk_fileref_does_file_exist) (frefid_t fref);

	void	(*glk_select) (event_t *event);
	void	(*glk_select_poll) (event_t *event);

	void	(*glk_request_timer_events) (glui32 millisecs);

	void	(*glk_request_line_event) (winid_t win, char *buf,
				glui32 maxlen, glui32 initlen);
	void	(*glk_request_char_event) (winid_t win);
	void	(*glk_request_mouse_event) (winid_t win);

	void	(*glk_cancel_line_event) (winid_t win, event_t *event);
	void	(*glk_cancel_char_event) (winid_t win);
	void	(*glk_cancel_mouse_event) (winid_t win);

	glui32	(*glk_image_draw) (winid_t win, glui32 image, glsi32 val1,
				glsi32 val2);
	glui32	(*glk_image_draw_scaled) (winid_t win, glui32 image,
				glsi32 val1, glsi32 val2, glui32 width,
				glui32 height);
	glui32	(*glk_image_get_info) (glui32 image, glui32 *width,
				glui32 *height);

	void	(*glk_window_flow_break) (winid_t win);

	void	(*glk_window_erase_rect) (winid_t win,
				glsi32 left, glsi32 top, glui32 width,
				glui32 height);
	void	(*glk_window_fill_rect) (winid_t win, glui32 color,
				glsi32 left, glsi32 top, glui32 width,
				glui32 height);
	void	(*glk_window_set_background_color) (winid_t win, glui32 color);

	schanid_t	(*glk_schannel_create) (glui32 rock);
	void		(*glk_schannel_destroy) (schanid_t chan);
	schanid_t	(*glk_schannel_iterate) (schanid_t chan,
				glui32 *rockptr);
	glui32	(*glk_schannel_get_rock) (schanid_t chan);

	glui32	(*glk_schannel_play) (schanid_t chan, glui32 snd);
	glui32	(*glk_schannel_play_ext) (schanid_t chan, glui32 snd,
				glui32 repeats, glui32 notify);
	void	(*glk_schannel_stop) (schanid_t chan);
	void	(*glk_schannel_set_volume) (schanid_t chan, glui32 vol);

	void	(*glk_sound_load_hint) (glui32 snd, glui32 flag);

	void	(*glk_set_hyperlink) (glui32 linkval);
	void	(*glk_set_hyperlink_stream) (strid_t str, glui32 linkval);
	void	(*glk_request_hyperlink_event) (winid_t win);
	void	(*glk_cancel_hyperlink_event) (winid_t win);

	void	(*glkunix_set_base_file) (char *filename);
	strid_t	(*glkunix_stream_open_pathname) (char *pathname,
				glui32 textmode, glui32 rock);

	/* Dispatch-layer Glk functions. */
	void	(*gidispatch_set_object_registry) (
			gidispatch_rock_t (*regi)(void *obj, glui32 objclass),
			void (*unregi)(void *obj, glui32 objclass,
						gidispatch_rock_t objrock));
	gidispatch_rock_t
		(*gidispatch_get_objrock) (void *obj, glui32 objclass);
	void	(*gidispatch_set_retained_registry) (
			gidispatch_rock_t (*regi)(void *array, glui32 len,
				char *typecode),
			void (*unregi)(void *array, glui32 len, char *typecode,
				gidispatch_rock_t objrock));

	void	(*gidispatch_call) (glui32 funcnum, glui32 numargs,
				gluniversal_t *arglist);
	char	*(*gidispatch_prototype) (glui32 funcnum);
	glui32	(*gidispatch_count_classes) (void);
	glui32	(*gidispatch_count_intconst) (void);
	gidispatch_intconst_t
		*(*gidispatch_get_intconst) (glui32 index);
	glui32	(*gidispatch_count_functions) (void);
	gidispatch_function_t
		*(*gidispatch_get_function) (glui32 index);
	gidispatch_function_t
		*(*gidispatch_get_function_by_id) (glui32 id);

	/* Blorb-layer Glk functions. */
	giblorb_err_t	(*giblorb_create_map) (strid_t file,
				giblorb_map_t **newmap);
	giblorb_err_t	(*giblorb_destroy_map) (giblorb_map_t *map);

	giblorb_err_t	(*giblorb_load_chunk_by_type) (giblorb_map_t *map,
				glui32 method, giblorb_result_t *res,
				glui32 chunktype, glui32 count);
	giblorb_err_t	(*giblorb_load_chunk_by_number) (giblorb_map_t *map,
				glui32 method, giblorb_result_t *res,
				glui32 chunknum);
	giblorb_err_t	(*giblorb_unload_chunk) (giblorb_map_t *map,
				glui32 chunknum);

	giblorb_err_t	(*giblorb_load_resource) (giblorb_map_t *map,
				glui32 method, giblorb_result_t *res,
				glui32 usage, glui32 resnum);
	giblorb_err_t	(*giblorb_count_resources) (giblorb_map_t *map,
				glui32 usage, glui32 *num, glui32 *min,
				glui32 *max);

	giblorb_err_t	(*giblorb_set_resource_map) (strid_t file);
	giblorb_map_t	*(*giblorb_get_resource_map)(void);

	/* Finally, a function we use for our own nefarious purposes. */
	strid_t		(*gli_stream_open_pathname) (char *pathname,
				int textmode, glui32 rock);
};


/*
 * Prototypes for Glk interception functions.  Instead of passing the
 * real Glk function addresses in a Glk interface, we can intercept them
 * with IFP versions.  At some point, these versions might do things like
 * demultiplexing by session ID or plugin.  For now, they're pass-through.
 */
extern void	ifp_glk_exit (void);
extern void	ifp_glk_set_interrupt_handler (void (*func)(void));
extern void	ifp_glk_tick (void);
extern glui32	ifp_glk_gestalt (glui32 sel, glui32 val);
extern glui32	ifp_glk_gestalt_ext (glui32 sel, glui32 val, glui32 *arr,
				glui32 arrlen);
extern unsigned char
		ifp_glk_char_to_lower (unsigned char ch);
extern unsigned char
		ifp_glk_char_to_upper (unsigned char ch);
extern winid_t	ifp_glk_window_get_root (void);
extern winid_t	ifp_glk_window_open (winid_t split, glui32 method, glui32 size,
				glui32 wintype, glui32 rock);
extern void	ifp_glk_window_close (winid_t win, stream_result_t *result);
extern void	ifp_glk_window_get_size (winid_t win, glui32 *widthptr,
			glui32 *heightptr);
extern void	ifp_glk_window_set_arrangement (winid_t win, glui32 method,
			glui32 size, winid_t keywin);
extern void	ifp_glk_window_get_arrangement (winid_t win, glui32 *methodptr,
			glui32 *sizeptr, winid_t *keywinptr);
extern winid_t	ifp_glk_window_iterate (winid_t win, glui32 *rockptr);
extern glui32	ifp_glk_window_get_rock (winid_t win);
extern glui32	ifp_glk_window_get_type (winid_t win);
extern winid_t	ifp_glk_window_get_parent (winid_t win);
extern winid_t	ifp_glk_window_get_sibling (winid_t win);
extern void	ifp_glk_window_clear (winid_t win);
extern void	ifp_glk_window_move_cursor (winid_t win, glui32 xpos,
			glui32 ypos);

extern strid_t	ifp_glk_window_get_stream (winid_t win);
extern void	ifp_glk_window_set_echo_stream (winid_t win, strid_t str);
extern strid_t	ifp_glk_window_get_echo_stream (winid_t win);
extern void	ifp_glk_set_window (winid_t win);

extern strid_t	ifp_glk_stream_open_file (frefid_t fileref, glui32 fmode,
			glui32 rock);
extern strid_t	ifp_glk_stream_open_memory (char *buf, glui32 buflen,
			glui32 fmode, glui32 rock);
extern void	ifp_glk_stream_close (strid_t str, stream_result_t *result);
extern strid_t	ifp_glk_stream_iterate (strid_t str, glui32 *rockptr);
extern glui32	ifp_glk_stream_get_rock (strid_t str);
extern void	ifp_glk_stream_set_position (strid_t str, glsi32 pos,
			glui32 seekmode);
extern glui32	ifp_glk_stream_get_position (strid_t str);
extern void	ifp_glk_stream_set_current (strid_t str);
extern strid_t	ifp_glk_stream_get_current (void);

extern void	ifp_glk_put_char (unsigned char ch);
extern void	ifp_glk_put_char_stream (strid_t str, unsigned char ch);
extern void	ifp_glk_put_string (char *s);
extern void	ifp_glk_put_string_stream (strid_t str, char *s);
extern void	ifp_glk_put_buffer (char *buf, glui32 len);
extern void	ifp_glk_put_buffer_stream (strid_t str, char *buf, glui32 len);
extern void	ifp_glk_set_style (glui32 styl);
extern void	ifp_glk_set_style_stream (strid_t str, glui32 styl);

extern glsi32	ifp_glk_get_char_stream (strid_t str);
extern glui32	ifp_glk_get_line_stream (strid_t str, char *buf, glui32 len);
extern glui32	ifp_glk_get_buffer_stream (strid_t str, char *buf, glui32 len);

extern void	ifp_glk_stylehint_set (glui32 wintype, glui32 styl, glui32 hint,
			glsi32 val);
extern void	ifp_glk_stylehint_clear (glui32 wintype, glui32 styl,
			glui32 hint);
extern glui32	ifp_glk_style_distinguish (winid_t win, glui32 styl1,
			glui32 styl2);
extern glui32	ifp_glk_style_measure (winid_t win, glui32 styl, glui32 hint,
			glui32 *result);

extern frefid_t	ifp_glk_fileref_create_temp (glui32 usage, glui32 rock);
extern frefid_t	ifp_glk_fileref_create_by_name (glui32 usage, char *name,
			glui32 rock);
extern frefid_t	ifp_glk_fileref_create_by_prompt (glui32 usage,
			glui32 fmode, glui32 rock);
extern frefid_t	ifp_glk_fileref_create_from_fileref (glui32 usage,
			frefid_t fref, glui32 rock);
extern void		ifp_glk_fileref_destroy (frefid_t fref);
extern frefid_t	ifp_glk_fileref_iterate (frefid_t fref, glui32 *rockptr);
extern glui32	ifp_glk_fileref_get_rock (frefid_t fref);
extern void	ifp_glk_fileref_delete_file (frefid_t fref);
extern glui32	ifp_glk_fileref_does_file_exist (frefid_t fref);

extern void	ifp_glk_select (event_t *event);
extern void	ifp_glk_select_poll (event_t *event);

extern void	ifp_glk_request_timer_events (glui32 millisecs);

extern void	ifp_glk_request_line_event (winid_t win, char *buf,
			glui32 maxlen, glui32 initlen);
extern void	ifp_glk_request_char_event (winid_t win);
extern void	ifp_glk_request_mouse_event (winid_t win);

extern void	ifp_glk_cancel_line_event (winid_t win, event_t *event);
extern void	ifp_glk_cancel_char_event (winid_t win);
extern void	ifp_glk_cancel_mouse_event (winid_t win);

extern glui32	ifp_glk_image_draw (winid_t win, glui32 image, glsi32 val1,
			glsi32 val2);
extern glui32	ifp_glk_image_draw_scaled (winid_t win, glui32 image,
			glsi32 val1, glsi32 val2, glui32 width,
			glui32 height);
extern glui32	ifp_glk_image_get_info (glui32 image, glui32 *width,
			glui32 *height);

extern void	ifp_glk_window_flow_break (winid_t win);

extern void	ifp_glk_window_erase_rect (winid_t win,
			glsi32 left, glsi32 top, glui32 width,
			glui32 height);
extern void	ifp_glk_window_fill_rect (winid_t win, glui32 color,
			glsi32 left, glsi32 top, glui32 width,
			glui32 height);
extern void	ifp_glk_window_set_background_color (winid_t win, glui32 color);

extern schanid_t
		ifp_glk_schannel_create (glui32 rock);
extern void	ifp_glk_schannel_destroy (schanid_t chan);
extern schanid_t
		ifp_glk_schannel_iterate (schanid_t chan,
			glui32 *rockptr);
extern glui32	ifp_glk_schannel_get_rock (schanid_t chan);

extern glui32	ifp_glk_schannel_play (schanid_t chan, glui32 snd);
extern glui32	ifp_glk_schannel_play_ext (schanid_t chan, glui32 snd,
			glui32 repeats, glui32 notify);
extern void	ifp_glk_schannel_stop (schanid_t chan);
extern void	ifp_glk_schannel_set_volume (schanid_t chan, glui32 vol);

extern void	ifp_glk_sound_load_hint (glui32 snd, glui32 flag);

extern void	ifp_glk_set_hyperlink (glui32 linkval);
extern void	ifp_glk_set_hyperlink_stream (strid_t str, glui32 linkval);
extern void	ifp_glk_request_hyperlink_event (winid_t win);
extern void	ifp_glk_cancel_hyperlink_event (winid_t win);

extern void	ifp_glkunix_set_base_file (char *filename);
extern strid_t	ifp_glkunix_stream_open_pathname (char *pathname,
			glui32 textmode, glui32 rock);

extern void	ifp_gidispatch_set_object_registry (
		gidispatch_rock_t (*regi)(void *obj, glui32 objclass),
		void (*unregi)(void *obj, glui32 objclass,
					gidispatch_rock_t objrock));
extern gidispatch_rock_t
		ifp_gidispatch_get_objrock (void *obj, glui32 objclass);
extern void	ifp_gidispatch_set_retained_registry (
		gidispatch_rock_t (*regi)(void *array, glui32 len,
			char *typecode),
		void (*unregi)(void *array, glui32 len, char *typecode,
			gidispatch_rock_t objrock));

extern void	ifp_gidispatch_call (glui32 funcnum, glui32 numargs,
			gluniversal_t *arglist);
extern char	*ifp_gidispatch_prototype (glui32 funcnum);
extern glui32	ifp_gidispatch_count_classes (void);
extern glui32	ifp_gidispatch_count_intconst (void);
extern gidispatch_intconst_t
		*ifp_gidispatch_get_intconst (glui32 index);
extern glui32	ifp_gidispatch_count_functions (void);
extern gidispatch_function_t
		*ifp_gidispatch_get_function (glui32 index);
extern gidispatch_function_t
		*ifp_gidispatch_get_function_by_id (glui32 id);

extern giblorb_err_t
		ifp_giblorb_create_map (strid_t file,
			giblorb_map_t **newmap);
extern giblorb_err_t
		ifp_giblorb_destroy_map (giblorb_map_t *map);

extern giblorb_err_t
		ifp_giblorb_load_chunk_by_type (giblorb_map_t *map,
			glui32 method, giblorb_result_t *res,
			glui32 chunktype, glui32 count);
extern giblorb_err_t
		ifp_giblorb_load_chunk_by_number (giblorb_map_t *map,
			glui32 method, giblorb_result_t *res,
			glui32 chunknum);
extern giblorb_err_t
		ifp_giblorb_unload_chunk (giblorb_map_t *map,
			glui32 chunknum);

extern giblorb_err_t
		ifp_giblorb_load_resource (giblorb_map_t *map,
			glui32 method, giblorb_result_t *res,
			glui32 usage, glui32 resnum);
extern giblorb_err_t
		ifp_giblorb_count_resources (giblorb_map_t *map,
			glui32 usage, glui32 *num, glui32 *min,
			glui32 *max);

extern giblorb_err_t
		ifp_giblorb_set_resource_map (strid_t file);
extern giblorb_map_t
		*ifp_giblorb_get_resource_map (void);

extern strid_t	ifp_gli_stream_open_pathname (char *pathname,
			int textmode, glui32 rock);

/*
 * Additional IFP prototypes for control interface functions.
 */
struct ifp_prefs;
extern int	ifpi_attach_glk_interface (ifp_glk_interfaceref_t interface);
extern ifp_glk_interfaceref_t
		ifpi_retrieve_glk_interface (void);
extern int	ifpi_attach_libc_interface (ifp_libc_interfaceref_t interface);
extern ifp_libc_interfaceref_t
		ifpi_retrieve_libc_interface (void);
extern void	ifpi_chain_set_plugin_self (ifp_pluginref_t plugin);
extern ifp_pluginref_t
		ifpi_chain_return_plugin (void);
extern void	ifpi_chain_accept_preferences (struct ifp_prefs *prefs_list);
extern int	ifpi_glkunix_startup_code (glkunix_startup_t *startdata);
extern void	ifpi_glk_main (void);
extern void	ifpi_force_link (void);

/* Internal-only interfaces. */
extern void	ifp_self_set_plugin (ifp_pluginref_t plugin);
extern ifp_pluginref_t
		ifp_self (void);
extern int	ifp_self_inside_plugin (void);
extern void	ifp_chain_set_chained_plugin (ifp_pluginref_t plugin);
extern ifp_pluginref_t
		ifp_chain_get_chained_plugin (void);
extern void	ifp_pref_use_foreign_data (struct ifp_prefs *prefs_list);
extern struct ifp_prefs
		*ifp_pref_get_local_data (void);
extern char	*ifp_cache_find_entry (char *url_path);
extern void	ifp_cache_release_entry (char *url_path);
extern void	ifp_cache_remove_entry (char *url_path);
extern int	ifp_cache_add_entry (char *url_path, char *data_file);
extern void	ifp_http_sigio_handler (void);
extern void	ifp_http_cancel_download (void);
extern int	ifp_http_download (int tofd, char *host,
			int port, char *document, int *progress, int *status);
extern void	ifp_ftp_sigio_handler (void);
extern void	ifp_ftp_cancel_download (void);
extern int	ifp_ftp_download (int tofd, char *host,
			int port, char *document, int *progress, int *status);
extern void	ifp_register_finalizer (void (*finalizer) (void));
extern strid_t	ifp_glkstream_open_pathname (char *pathname,
			glui32 textmode, glui32 rock);
extern void	ifp_glkstream_close (strid_t glk_stream,
			stream_result_t *result);
extern void	*ifp_port_dlopen (const char *filename);
extern char	*ifp_port_dlerror (void);
extern void	*ifp_port_dlsym (void *handle, char *symbol);
extern int	ifp_port_dlclose (void *handle);
extern void	*ifp_port_malloc (int size);
extern void	*ifp_port_realloc (void *ptr, int size);
extern void	ifp_port_free (void *pointer);
extern int	ifp_port_is_filename_shared_object (const char *filename);
extern void	ifp_port_abort (void);

#endif /* IFP_HEADER_ONLY */
#endif /* IFP_INTERNAL_H */
